from .version import INFESTOR_VERSION

from humbug.consent import HumbugConsent, environment_variable_opt_in, yes
from humbug.report import HumbugReporter

INFESTOR_REPORTING_TOKEN = "6ec64442-40e3-41ff-afe3-d818c023cd41"

infestor_consent = HumbugConsent(
    environment_variable_opt_in("INFESTOR_REPORTING_ENABLED", yes)
)

infestor_reporter = HumbugReporter(
    name=f"infestor",
    consent=infestor_consent,
    bugout_token=INFESTOR_REPORTING_TOKEN,
)

infestor_tags = [f"version:{INFESTOR_VERSION}"]



from pydantic import BaseModel
from typing import List


class ReporterCall(BaseModel):
    call_type: str
    lineno: int
    scope_stack: str


class ReporterDecorator(BaseModel):
    decorator_type: str
    lineno: int
    scope_stack: str


class ReporterDecoratorCandidate(BaseModel):
    scope_stack: str
    lineno: int



import logging
import os
from typing import cast, Dict, List, Optional, Sequence
from . import models
from .errors import *
from .manager import (
    PackageFileManager,
    DECORATOR_TYPE_RECORD_ERRORS,
    DECORATOR_TYPE_RECORD_CALL,
    CALL_TYPE_SETUP_EXCEPTHOOK,
    CALL_TYPE_SYSTEM_REPORT,
)
from .config import (
    default_config_file,
    load_config,
    save_config,
    python_root_relative_to_repository_root,
)

DEFAULT_REPORTER_FILENAME = "report.py"
REPORTER_FILE_TEMPLATE: Optional[str] = None
TEMPLATE_FILEPATH = os.path.join(os.path.dirname(__file__), "report.py.template")

try:
    with open(TEMPLATE_FILEPATH, "r") as ifp:
        REPORTER_FILE_TEMPLATE = ifp.read()
except Exception as e:
    logging.warn(f"WARNING: Could not load reporter template from {TEMPLATE_FILEPATH}:")
    logging.warn(e)


def python_files(repository: str) -> Sequence[str]:
    results: List[str] = []
    if os.path.isfile(repository):
        return [repository]

    for dirpath, _, filenames in os.walk(repository, topdown=True):
        results.extend(
            [
                os.path.join(dirpath, filename)
                for filename in filenames
                if filename.endswith(".py")
            ]
        )

    return results


def list_calls(
    call_type: str,
    repository: str,
    candidate_files: Optional[Sequence[str]] = None,
) -> Dict[str, List[models.ReporterCall]]:
    """
    Args:
    0. call_type - Type of call to list in the given package (e.g. "system_report", "setup_excepthook", etc.)
    1. repository - Path to repository in which Infestor has been set up
    2. candidate_files - Optional list of files to restrict analysis to
    """
    results: Dict[str, List[models.ReporterCall]] = {}
    if candidate_files is None:
        candidate_files = python_files(repository)

    for filepath in candidate_files:
        package_file_manager = PackageFileManager(repository, filepath)
        calls = package_file_manager.get_calls(call_type)
        if calls:
            results[filepath] = calls

    return results


def add_call(
    call_type: str,
    repository: str,
    submodule_path: Optional[str] = None,
) -> None:
    """
    Args:
    0. call_type - Type of call to add to the given package (e.g. "system_report", "setup_excepthook", etc.)
    1. repository - Path to repository in which Infestor has been set up
    2. submodule_path: Path (relative to python_root) of file in which we want to add a sytem_report
    """
    # Determine the file to which we should add the call
    target_file = submodule_path
    if target_file is None:
        target_file = repository
        if os.path.isdir(target_file):
            target_file = os.path.join(target_file, "__init__.py")

    if not os.path.exists(target_file):
        with open(target_file, "w") as ofp:
            ofp.write("")

    package_file_manager = PackageFileManager(repository, target_file)
    package_file_manager.add_call(call_type)
    package_file_manager.write_to_file()


def remove_calls(
    call_type: str,
    repository: str,
    submodule_path: Optional[str] = None,
) -> None:
    """
    Args:
    0. call_type - Type of call to remove from the given package (e.g. "system_report", "setup_excepthook", etc.)
    1. repository - Path to repository in which Infestor has been set up
    2. submodule_path: Path (relative to python_root) of file in which we want to add a sytem_report
    """
    candidate_files: Sequence[str] = []

    if submodule_path is None:
        candidate_files = python_files(repository)
    else:
        candidate_files = [submodule_path]

    for candidate_file in candidate_files:
        package_file_manager = PackageFileManager(repository, candidate_file)
        package_file_manager.remove_call(call_type)
        package_file_manager.write_to_file()


def list_decorators(
    decorator_type: str,
    repository: str,
    candidate_files: Optional[Sequence[str]] = None,
) -> Dict[str, List[models.ReporterDecorator]]:
    """
    Args:
    0. decorator_type - Type of decorator to list in the given package (choices: "record_call", "record_error")
    1. repository - Path to repository in which Infestor has been set up
    2. python_root - Path (relative to repository) of Python package to work with (used to parse config)
    3. candidate_files - Optional list of files to restrict analysis to

    Returns a dictionary mapping file paths to functions defined in those files decorated by the given decorator_type method
    on a managed Humbug reporter.
    """
    results: Dict[str, List[models.ReporterDecorator]] = {}

    if candidate_files is None:
        candidate_files = python_files(repository)

    for candidate_file in candidate_files:
        package_file_manager = PackageFileManager(repository, candidate_file)
        decorators = package_file_manager.list_decorators(decorator_type)
        if decorators:
            results[candidate_file] = decorators

    return results


def decorator_candidates(
    decorator_type: str,
    repository: str,
    submodule_path: str,
) -> List[models.ReporterDecoratorCandidate]:
    """
    Args:
    0. decorator_type - Type of decorator to add to the given package (choices: "record_call", "record_error")
    1. repository - Path to repository in which Infestor has been set up
    2. submodule_path: Path (relative to python_root) of file in which we want to add a sytem_report

    Returns a list of tuples of the form:
    (
        <function name>,
        <function definition starting line number>
    )

    This list is the list of candidate function definitions that the user could decorate (i.e. the ones which
    do not already have a decorator of the given decorator_type).
    """
    package_file_manager = PackageFileManager(repository, submodule_path)
    return package_file_manager.decorator_candidates(decorator_type)


def add_decorators(
    decorator_type: str,
    repository: str,
    submodule_path: str,
    linenos: List[int],
) -> None:
    """
    Args:
    0. decorator_type - Type of decorator to add to the given package (choices: "record_call", "record_error")
    1. repository - Path to repository in which Infestor has been set up
    2. submodule_path: Path (relative to python_root) of file in which we want to add a sytem_report
    3. linenos: Line numbers where functions are defined that we wish to decorate
    """

    candidates = decorator_candidates(decorator_type, repository, submodule_path)

    candidate_linenos = []
    for candidate in candidates:
        candidate_linenos.append(candidate.lineno)

    for lineno in linenos:
        if lineno not in candidate_linenos:
            raise GenerateDecoratorError(
                f"Non-candidate source code: submodule_path={submodule_path}, lineno={lineno}"
            )

    package_file_manager = PackageFileManager(repository, submodule_path)
    package_file_manager.add_decorators(decorator_type, linenos)
    package_file_manager.write_to_file()


def remove_decorators(
    decorator_type: str,
    repository: str,
    submodule_path: str,
    linenos: List[int],
) -> None:
    """
    Args:
    0. decorator_type - Type of decorator to remove from the given package (choices: "record_call", "record_error")
    1. repository - Path to repository in which Infestor has been set up
    2. submodule_path: Path (relative to python_root) of file in which we want to add a sytem_report
    3. linenos: Line numbers where decorated functions are defined that we wish to undecorate
    """
    candidates_for_removal = list_decorators(
        decorator_type, repository, [submodule_path]
    ).get(submodule_path, [])

    candidate_linenos = []
    for candidate in candidates_for_removal:
        candidate_linenos.append(candidate.lineno)

    for lineno in linenos:
        if lineno not in candidate_linenos:
            raise GenerateDecoratorError(
                f"Could not undecorate invalid code at: submodule_path={submodule_path}, lineno={lineno}"
            )

    package_file_manager = PackageFileManager(repository, submodule_path)
    package_file_manager.remove_decorators(decorator_type, linenos)
    package_file_manager.write_to_file()


def add_reporter(
    repository: str,
    reporter_filepath: Optional[str] = None,
    force: bool = False,
) -> None:
    if REPORTER_FILE_TEMPLATE is None:
        raise GenerateReporterError("Could not load reporter template file")

    config_file = default_config_file(repository)
    configuration = load_config(config_file)

    if reporter_filepath is None:
        if configuration.reporter_filepath is not None:
            reporter_filepath = configuration.reporter_filepath
        else:
            reporter_filepath = DEFAULT_REPORTER_FILENAME
    else:
        if (
            configuration.reporter_filepath is not None
            and configuration.reporter_filepath != reporter_filepath
        ):
            raise GenerateReporterError(
                f"Configuration expects reporter to be set up at a different file than the one specified; specified={reporter_filepath}, expected={configuration.reporter_filepath}"
            )

    reporter_filepath_full = os.path.join(repository, reporter_filepath)
    if (not force) and os.path.exists(reporter_filepath_full):
        raise GenerateReporterError(
            f"Object already exists at desired reporter filepath: {reporter_filepath_full}"
        )

    if configuration.reporter_token is None:
        raise GenerateReporterError("No reporter token was specified in configuration")

    contents = REPORTER_FILE_TEMPLATE.format(
        project_name=configuration.project_name,
        reporter_token=configuration.reporter_token,
    )
    with open(reporter_filepath_full, "w") as ofp:
        ofp.write(contents)

    configuration.reporter_filepath = reporter_filepath
    save_config(config_file, configuration)



import os
import unittest

from . import config
from . import visitors
from .operations import add_reporter
from .testcase import InfestorTestCase


class TestReporterFileVisitor(InfestorTestCase):
    def setUp(self):
        super().setUp()
        add_reporter(self.package_dir)
        self.config = config.load_config(self.config_file)

    def test_visitor(self):
        visitor = visitors.ReporterFileVisitor()
        syntax_tree = visitor.syntax_tree(
            os.path.join(self.package_dir, self.config.reporter_filepath)
        )
        syntax_tree.visit(visitor)
        self.assertEqual(visitor.HumbugConsentImportedAs, "HumbugConsent")
        self.assertEqual(visitor.HumbugConsentInstantiatedAs, "consent")
        self.assertGreater(visitor.HumbugConsentImportedAt, 0)
        self.assertGreater(
            visitor.HumbugConsentInstantiatedAt, visitor.HumbugConsentImportedAt
        )
        self.assertEqual(visitor.HumbugReporterImportedAs, "HumbugReporter")
        self.assertEqual(visitor.HumbugReporterInstantiatedAs, "reporter")
        self.assertGreater(
            visitor.HumbugReporterImportedAt, visitor.HumbugConsentImportedAt
        )
        self.assertGreater(
            visitor.HumbugReporterInstantiatedAt, visitor.HumbugReporterImportedAt
        )
        self.assertGreater(
            visitor.HumbugReporterInstantiatedAt, visitor.HumbugConsentInstantiatedAt
        )
        self.assertEqual(
            visitor.HumbugReporterConsentArgument, visitor.HumbugConsentInstantiatedAs
        )
        self.assertNotEqual(visitor.HumbugReporterTokenArgument, "")


class TestPackageFileVisitor(unittest.TestCase):
    def setUp(self):
        self.source = ""


if __name__ == "__main__":
    unittest.main()



import enum
import glob
import os
import shutil
import sys
import tempfile
import unittest
import uuid

import pygit2

from . import commit
from . import config


class InfestorTestCase(unittest.TestCase):
    def setUp(self):
        self.repository = tempfile.mkdtemp()
        pygit2.init_repository(self.repository, False)

        self.fixtures_dir = os.path.join(os.path.dirname(__file__), "fixtures")

        script_basename = "a_script.py"
        script_file_fixture = os.path.join(self.fixtures_dir, script_basename)
        self.script_file = os.path.join(self.repository, script_basename)
        shutil.copyfile(script_file_fixture, self.script_file)

        package_basename = "a_package"
        package_dir_fixture = os.path.join(self.fixtures_dir, package_basename)
        self.package_dir = os.path.join(self.repository, package_basename)
        shutil.copytree(package_dir_fixture, self.package_dir)

        self.package_name = "a_package"

        self.reporter_token = str(uuid.uuid4())

        self.config = config.initialize(
            self.package_dir,
            self.package_name,
            reporter_token=self.reporter_token,
        )

        self.config_file = config.default_config_file(self.package_dir)

        package_files = [
            os.path.relpath(python_file, start=self.repository)
            for python_file in glob.glob(os.path.join(self.package_dir, "*.py"))
        ]
        commit.commit_files(
            self.repository,
            "refs/heads/master",
            [
                script_basename,
                *package_files,
                os.path.join(package_basename, config.CONFIG_FILENAME),
            ],
            "initial commit",
        )

    def tearDown(self) -> None:
        DEBUG = os.getenv("DEBUG")
        if DEBUG != "1":
            shutil.rmtree(self.repository)
        else:
            print(
                f"DEBUG=1: Retaining test directory - {self.repository}",
                file=sys.stderr,
            )



from typing import Tuple, List, Optional
import os
import libcst as cst
from . import visitors
from . import transformers
from .errors import *
from .config import (
    default_config_file,
    load_config,
    save_config,
    python_root_relative_to_repository_root,
)

# TODO(zomglings): Use an Enum here.
CALL_TYPE_SYSTEM_REPORT = "system_report"
CALL_TYPE_SETUP_EXCEPTHOOK = "setup_excepthook"

DECORATOR_TYPE_RECORD_CALL = "record_call"
DECORATOR_TYPE_RECORD_ERRORS = "record_errors"


def get_reporter_module_path(repository: str, submodule_path: str) -> Tuple[str, bool]:

    config_file = default_config_file(repository)
    configuration = load_config(config_file)
    if configuration is None:
        raise GenerateReporterError(
            f"Could not load configuration from file ({config_file})"
        )

    if configuration.reporter_filepath is None:
        raise GenerateReporterError(
            f"No reporter defined for project. Try running:\n\t$ infestor -r {repository} generate setup -o report.py"
        )
    reporter_filepath = os.path.join(repository, configuration.reporter_filepath)

    if not os.path.exists(submodule_path):
        raise GenerateReporterError(f"No file at submodule_path: {submodule_path}")

    path_to_reporter_file = os.path.relpath(
        os.path.join(repository, reporter_filepath),
        os.path.dirname(submodule_path),
    )
    path_components: List[str] = []
    current_path = path_to_reporter_file
    while current_path:
        current_path, base = os.path.split(current_path)
        if base == os.path.basename(reporter_filepath):
            base, _ = os.path.splitext(base)
        path_components = [base] + path_components
    # TODO: path components for /fixtures/a_script looks
    #  like ['..', 'report'], for /fixtures/a_package: ['..', '..', 'report']
    #  so am making '..' to '.' to make it work

    name: Optional[str] = None
    if not configuration.relative_imports:
        path_components = [os.path.basename(repository)] + path_components
        name = f"{os.path.basename(repository)}.{path_components[-1]}"
        # name = ".".join(path_components)

    else:
        name = "."
        for comp in path_components:
            if comp == "..":
                name += "."
            else:
                name += comp
        # name = "." + ".".join(path_components)

    return (name, configuration.relative_imports)


class PackageFileManager:
    def __init__(self, repository: str, filepath: str):
        self.filepath = filepath
        self.repository = repository
        self._load_file(filepath)

    def _load_file(self, filepath: str):
        self.reporter_module_path, self.relative_imports = get_reporter_module_path(
            self.repository, filepath
        )
        with open(filepath, "r") as ifp:
            file_source = ifp.read()
        self._visit(cst.parse_module(file_source))

    def _visit(self, module: cst.Module):
        self.syntax_tree = cst.metadata.MetadataWrapper(module)
        self.visitor = visitors.PackageFileVisitor(
            self.reporter_module_path, self.relative_imports
        )
        self.syntax_tree.visit(self.visitor)

    def get_code(self):
        return self.syntax_tree.module.code

    def write_to_file(self):
        with open(self.filepath, "w") as ofp:
            ofp.write(self.get_code())

    def is_reporter_imported(self) -> bool:
        return (
            self.visitor.ReporterImportedAt != -1
            and self.visitor.ReporterImportedAs != ""
        )

    def ensure_import_reporter(self):
        if not self.is_reporter_imported():
            self.add_reporter_import()

    def add_reporter_import(self) -> None:
        if self.is_reporter_imported():
            return
        transformer = transformers.ImportReporterTransformer(self.reporter_module_path)
        modified_tree = self.syntax_tree.visit(transformer)
        self._visit(modified_tree)

    def get_calls(self, call_type):
        return self.visitor.calls.get(call_type, [])

    def add_call(self, call_type):
        if self.get_calls(call_type):
            return
        self.ensure_import_reporter()
        transformer = transformers.ReporterCallsAdderTransformer(
            self.visitor.ReporterImportedAs, call_type
        )
        modified_tree = self.syntax_tree.visit(transformer)
        self._visit(modified_tree)

    def remove_call(self, call_type: str):
        transformer = transformers.ReporterCallsRemoverTransformer(
            self.visitor.ReporterImportedAs, call_type
        )

        modified_tree = self.syntax_tree.visit(transformer)
        self._visit(modified_tree)

    def list_decorators(self, decorator_type: str):
        return self.visitor.decorators.get(decorator_type, [])

    def decorator_candidates(self, decorator_type: str):
        decorator_candidates_visitor = visitors.DecoratorCandidatesVisitor(
            self.visitor.ReporterImportedAs, decorator_type
        )
        self.syntax_tree.visit(decorator_candidates_visitor)
        return decorator_candidates_visitor.decorator_candidates

    def add_decorators(self, decorator_type: str, linenos: List[int]):
        func_linenos = linenos
        if not self.is_reporter_imported():
            # Reporter not importer, we need to add reporter
            self.add_reporter_import()
            # Since we added reporter, source has changed and
            # all linenos need to be increased by 1
            func_linenos = [x + 1 for x in func_linenos]

        transformer = transformers.DecoratorsAdderTransformer(
            self.visitor.ReporterImportedAs, decorator_type, func_linenos
        )
        modified_tree = self.syntax_tree.visit(transformer)
        self._visit(modified_tree)
        if decorator_type == DECORATOR_TYPE_RECORD_ERRORS:
            # We've added decorator which shifts linenos
            try_except_transformer = transformers.TryExceptAdderTransformer(
                self.visitor.ReporterImportedAs, [x + 1 for x in func_linenos]
            )
            modified_tree = self.syntax_tree.visit(try_except_transformer)
            self._visit(modified_tree)

    def remove_decorators(self, decorator_type: str, linenos: List[int]):
        transformer = transformers.DecoratorsRemoverTransformer(
            self.visitor.ReporterImportedAs, decorator_type, linenos
        )
        modified_tree = self.syntax_tree.visit(transformer)
        self._visit(modified_tree)

        if decorator_type == DECORATOR_TYPE_RECORD_ERRORS:
            try_except_transformer = transformers.TryExceptRemoverTransformer(
                self.visitor.ReporterImportedAs, [x - 1 for x in linenos]
            )
            modified_tree = self.syntax_tree.visit(try_except_transformer)
            self._visit(modified_tree)



import json
import os
import unittest
import libcst as cst
from . import operations
from . import visitors
from .testcase import InfestorTestCase


class TestSetupReporter(InfestorTestCase):
    def test_add_reporter_for_package(self):
        # Steps:
        # 1. Initialize infestor config in given repository, with Python package as target.
        # 2. Add reporter to package
        # 3. Check that reporter file did not exist before reporter was added, but that it DOES exist afterwards.
        # 4. Check that structure of reporter file contains:
        #    a. Import of HumbugConsent from humbug.consent
        #    b. Import of HumbugReporter from humbug.reporter
        #    c. Instantiation of HumbugConsent into a variable (store the name of this variable)
        #    d. Instantiation of HumbugReporter with consent variable as an argument
        #    e. Instantiation of HumbugReporter with the configured token as an argument
        with open(self.config_file, "r") as ifp:
            infestor_json_old = json.load(ifp)
        self.assertIsNone(infestor_json_old["reporter_filepath"])

        reporter_filepath = os.path.join(self.package_dir, "report.py")
        self.assertFalse(os.path.exists(reporter_filepath))

        operations.add_reporter(self.package_dir)

        with open(self.config_file, "r") as ifp:
            infestor_json_new = json.load(ifp)
        self.assertEqual(infestor_json_new["reporter_filepath"], "report.py")
        self.assertTrue(os.path.exists(reporter_filepath))

        reporter_visitor = visitors.ReporterFileVisitor()
        reporter_syntax_tree = reporter_visitor.syntax_tree(reporter_filepath)

        reporter_syntax_tree.visit(reporter_visitor)

        self.assertEqual(reporter_visitor.HumbugConsentImportedAs, "HumbugConsent")
        self.assertLess(
            reporter_visitor.HumbugConsentImportedAt,
            reporter_visitor.HumbugConsentInstantiatedAt,
        )
        self.assertLess(
            reporter_visitor.HumbugReporterImportedAt,
            reporter_visitor.HumbugReporterInstantiatedAt,
        )
        self.assertEqual(
            reporter_visitor.HumbugReporterConsentArgument,
            reporter_visitor.HumbugConsentInstantiatedAs,
        )
        self.assertEqual(
            reporter_visitor.HumbugReporterTokenArgument,
            f"\"{infestor_json_new['reporter_token']}\"",
        )

    def test_system_report_add_with_no_reporter_added(self):
        with self.assertRaises(operations.GenerateReporterError):
            operations.add_call(
                operations.CALL_TYPE_SYSTEM_REPORT,
                self.package_dir,
            )

    def test_list_system_reports_for_package_with_no_system_reports(self):
        operations.add_reporter(self.package_dir)
        results = operations.list_calls(
            operations.CALL_TYPE_SYSTEM_REPORT,
            self.package_dir,
        )
        self.assertDictEqual(results, {})

    def test_decorator_list_with_no_reporter_decorators(self):
        operations.add_reporter(self.package_dir)
        results = operations.list_decorators(
            operations.DECORATOR_TYPE_RECORD_ERRORS, self.package_dir
        )
        self.assertDictEqual(results, {})

    def test_decorator_add_remove(self):
        operations.add_reporter(self.package_dir)
        target_file = os.path.join(self.package_dir, "cli.py")
        candidates = operations.decorator_candidates(
            operations.DECORATOR_TYPE_RECORD_ERRORS, self.package_dir, target_file
        )
        self.assertNotEqual(len(candidates), 0, "Failed to find decorator candidates")
        linenos = []
        for candidate in candidates:
            linenos.append(candidate.lineno)

        operations.add_decorators(
            operations.DECORATOR_TYPE_RECORD_ERRORS,
            self.package_dir,
            target_file,
            linenos,
        )

        new_candidates = operations.decorator_candidates(
            operations.DECORATOR_TYPE_RECORD_ERRORS, self.package_dir, target_file
        )

        self.assertEqual(len(new_candidates), 0, "Failed to decorate all candidates")

        decorators = operations.list_decorators(
            operations.DECORATOR_TYPE_RECORD_ERRORS, self.package_dir, [target_file]
        )

        self.assertNotEqual(decorators, {}, "Failed to list decorators")

        self.assertEqual(
            len(decorators[target_file]),
            len(candidates),
            "Failed to list all decorators",
        )

        linenos = []
        for decorator in decorators[target_file]:
            linenos.append(decorator.lineno)

        operations.remove_decorators(
            operations.DECORATOR_TYPE_RECORD_ERRORS,
            self.package_dir,
            target_file,
            linenos,
        )

        decorators = operations.list_decorators(
            operations.DECORATOR_TYPE_RECORD_ERRORS, self.package_dir, [target_file]
        )

        self.assertEqual(decorators, {}, "Failed to remove decorators")

    def test_system_report_add(self):
        operations.add_reporter(self.package_dir)
        operations.add_call(operations.CALL_TYPE_SYSTEM_REPORT, self.package_dir)

        target_file = os.path.join(self.package_dir, "__init__.py")

        source = ""
        with open(target_file, "r") as ifp:
            for line in ifp:
                source += line

        source_tree = cst.metadata.MetadataWrapper(cst.parse_module(source))
        visitor = visitors.PackageFileVisitor(self.package_name + ".report", False)
        source_tree.visit(visitor)

        self.assertNotEqual(visitor.ReporterImportedAt, -1, "reporter not imported")

        self.assertNotEqual(
            len(visitor.calls.get("system_report")), 0, "system_call not called"
        )

        self.assertEqual(
            visitor.last_import_lineno,
            visitor.ReporterImportedAt,
            "reporter is not last import",
        )
        self.assertTrue(
            visitor.ReporterCorrectlyImported,
            "reporter is not imported right after last naked import",
        )
        self.assertEqual(
            visitor.ReporterImportedAt,
            visitor.calls.get("system_report")[0].lineno - 1,
            "system_call is not called right after import",
        )

    def test_system_report_remove(self):
        operations.add_reporter(self.package_dir)
        operations.add_call(operations.CALL_TYPE_SYSTEM_REPORT, self.package_dir)
        calls = operations.list_calls(
            operations.CALL_TYPE_SYSTEM_REPORT, self.package_dir
        )
        self.assertNotEqual(calls, {}, "Failed to add system_report call")

        operations.remove_calls(operations.CALL_TYPE_SYSTEM_REPORT, self.package_dir)
        calls = operations.list_calls(
            operations.CALL_TYPE_SYSTEM_REPORT, self.package_dir
        )
        self.assertEqual(calls, {}, "Failed to remove system_report call")


if __name__ == "__main__":
    unittest.main()



"""
Command line interface for the Humbug infestor.
"""
import argparse
import os
from typing import Callable, Tuple

from . import config, operations

CLIHandler = Callable[[argparse.Namespace], None]


def handle_config_init(args: argparse.Namespace) -> None:
    config.initialize(
        args.repository,
        args.name,
        args.relative_imports,
        args.reporter_token,
    )
    print("Infestor has been initialized for your repository.\n")

    print(
        "The Bugout team would like to collect crash reports as well as some basic, anonymous "
        "information about your system when you run infestor.\nThis will help us improve the "
        "infestor experience for everyone.\nWe request that you opt into reporting by setting the "
        "environment variable:\n"
        "\tINFESTOR_REPORTING_ENABLED=yes\n"
    )
    print("In bash or zsh:\n\t$ export INFESTOR_REPORTING_ENABLED=yes")
    print("In fish:\n\t$ set -x INFESTOR_REPORTING_ENABLED yes")
    print("On Windows (cmd or powershell):\n\t$ set INFESTOR_REPORTING_ENABLED=yes\n")

    if args.reporter_token is None:
        print(
            "It looks like you have not configured Infestor with a Bugout reporter token."
        )
        print(
            "Generate a reporter token by adding an integration to your team at https://bugout.dev/account/teams."
        )
        print("Once you have generated a token, run:")
        print("\t$ infestor token <token>\n")


def handle_config_validate(args: argparse.Namespace) -> None:
    config_file = config.default_config_file(args.repository)
    config.load_config(config_file, print_warnings=True)


def handle_config_token(args: argparse.Namespace) -> None:
    config_file = config.default_config_file(args.repository)
    config_object = config.set_reporter_token(
        config_file,
        args.token,
    )

    if config_object.reporter_filepath is not None:
        operations.add_reporter(
            args.repository,
            config_object.reporter_filepath,
            force=True,
        )

    print(config_object)


def handle_reporter_add(args: argparse.Namespace) -> None:
    operations.add_reporter(args.repository, args.reporter_filepath, args.force)


def handle_report_all(args: argparse.Namespace) -> None:
    operations.add_call(operations.CALL_TYPE_SYSTEM_REPORT, args.repository)
    operations.add_call(operations.CALL_TYPE_SETUP_EXCEPTHOOK, args.repository)

    files = operations.python_files(args.repository)
    for file in files:
        candidates = operations.decorator_candidates(
            operations.DECORATOR_TYPE_RECORD_ERRORS, args.repository, file
        )
        candidate_linenos = []
        for candidate in candidates:
            candidate_linenos.append(candidate.lineno)
        if candidate_linenos:
            operations.add_decorators(
                operations.DECORATOR_TYPE_RECORD_ERRORS,
                args.repository,
                file,
                candidate_linenos,
            )
    for file in files:
        candidates = operations.decorator_candidates(
            operations.DECORATOR_TYPE_RECORD_CALL, args.repository, file
        )
        candidate_linenos = []
        for candidate in candidates:
            candidate_linenos.append(candidate.lineno)
        if candidate_linenos:
            operations.add_decorators(
                operations.DECORATOR_TYPE_RECORD_CALL,
                args.repository,
                file,
                candidate_linenos,
            )


def generate_call_handlers(call_type: str) -> Tuple[CLIHandler, CLIHandler, CLIHandler]:
    """
    Returns a tuple of the form:
    (handle_list, handle_add, handle_remove)
    """

    def handle_list(args: argparse.Namespace) -> None:
        results = operations.list_calls(call_type, args.repository)
        for filepath, calls in results.items():
            print(f"Lines in {filepath}:")
            for call in calls:
                print(f"\t- (line {call.lineno}) {call.scope_stack}")

    def handle_add(args: argparse.Namespace) -> None:
        # TODO(zomglings): Is there a better way to check if an argparse.Namespace has a given member?
        if vars(args).get("submodule") is not None:
            operations.add_call(call_type, args.repository, args.submodule)
        else:
            operations.add_call(call_type, args.repository)

    def handle_remove(args: argparse.Namespace) -> None:
        # TODO(zomglings): Ditto
        if vars(args).get("submodule") is not None:
            operations.remove_calls(call_type, args.repository, args.submodule)
        else:
            operations.remove_calls(call_type, args.repository)

    return (handle_list, handle_add, handle_remove)


def generate_decorator_handlers(
    decorator_type: str,
) -> Tuple[CLIHandler, CLIHandler, CLIHandler, CLIHandler]:
    """
    Returns a tuple of the form:
    (handle_list, handle_candidates, handle_add, handle_remove)
    """

    def handle_list(args: argparse.Namespace) -> None:
        results = operations.list_decorators(decorator_type, args.repository)
        for filepath, decorators in results.items():
            print(f"Lines in {filepath}:")
            for decorator in decorators:
                print(f"\t- (line {decorator.lineno}) {decorator.scope_stack}")

    def handle_candidates(args: argparse.Namespace) -> None:
        results = operations.decorator_candidates(
            decorator_type, args.repository, args.submodule
        )
        print(f"You can add the {decorator_type} decorator to the following functions:")
        for candidate in results:
            print(f"\t- (line {candidate.lineno}) {candidate.scope_stack}")

    def handle_add(args: argparse.Namespace) -> None:
        operations.add_decorators(
            decorator_type,
            args.repository,
            args.submodule,
            args.lines,
        )

    def handle_remove(args: argparse.Namespace) -> None:
        operations.remove_decorators(
            decorator_type,
            args.repository,
            args.submodule,
            args.lines,
        )

    return (handle_list, handle_candidates, handle_add, handle_remove)


def generate_argument_parser() -> argparse.ArgumentParser:
    current_working_directory = os.getcwd()

    parser = argparse.ArgumentParser(
        description="Infestor: Manage Humbug instrumentation of your Python code base"
    )
    parser.set_defaults(func=lambda _: parser.print_help())
    subcommands = parser.add_subparsers()

    def populate_leaf_parser_with_common_args(
        leaf_parser: argparse.ArgumentParser,
    ) -> None:
        leaf_parser.add_argument(
            "-r",
            "--repository",
            default=current_working_directory,
            help=f"Path to git repository containing your code base (default: {current_working_directory})",
        )

    config_parser = subcommands.add_parser(
        "config", description="Manage infestor configuration"
    )
    config_parser.set_defaults(func=lambda _: config_parser.print_help())
    config_subcommands = config_parser.add_subparsers()

    config_init_parser = config_subcommands.add_parser(
        "init", description="Initialize an Infestor integration in a project"
    )
    populate_leaf_parser_with_common_args(config_init_parser)
    config_init_parser.add_argument(
        "-n",
        "--name",
        required=True,
        help="Name of project (to identify integration)",
    )
    config_init_parser.add_argument(
        "--relative-imports",
        action="store_true",
        help="Set this flags if infestor should add relative imports.",
    )
    config_init_parser.add_argument(
        "-t",
        "--reporter-token",
        default=None,
        help="Bugout reporter token. Get one by setting up an integration at https://bugout.dev/account/teams",
    )
    config_init_parser.set_defaults(func=handle_config_init)

    config_validate_parser = config_subcommands.add_parser(
        "validate", description="Validate an Infestor configuration"
    )
    populate_leaf_parser_with_common_args(config_validate_parser)
    config_validate_parser.set_defaults(func=handle_config_validate)

    config_token_parser = config_subcommands.add_parser(
        "token", description="Set a Humbug token for an Infestor integration"
    )
    populate_leaf_parser_with_common_args(config_token_parser)
    config_token_parser.add_argument(
        "token", help="Reporting token generated from https://bugout.dev/account/teams"
    )
    config_token_parser.set_defaults(func=handle_config_token)

    reporter_parser = subcommands.add_parser(
        "reporter", description="Manage Humbug reporters in a code base"
    )
    reporter_parser.set_defaults(func=lambda _: reporter_parser.print_help())
    reporter_subcommands = reporter_parser.add_subparsers()

    reporter_add_parser = reporter_subcommands.add_parser(
        "add", description="Adds a Humbug reporter to a Python package"
    )
    populate_leaf_parser_with_common_args(reporter_add_parser)
    reporter_add_parser.add_argument(
        "-o",
        "--reporter-filepath",
        required=False,
        default=operations.DEFAULT_REPORTER_FILENAME,
        help=f"Path (relative to Python root) at which we should set up the reporter integration (default: {operations.DEFAULT_REPORTER_FILENAME})",
    )
    reporter_add_parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="Set this flag if you want to overwrite the reporter file if it already exists",
    )
    reporter_add_parser.set_defaults(func=handle_reporter_add)

    system_report_parser = subcommands.add_parser(
        "system-report", description="Manage Humbug system reporting in a code base"
    )
    system_report_parser.set_defaults(func=lambda _: system_report_parser.print_help())
    system_report_subcommands = system_report_parser.add_subparsers()

    (
        handle_system_report_list,
        handle_system_report_add,
        handle_system_report_remove,
    ) = generate_call_handlers(operations.CALL_TYPE_SYSTEM_REPORT)

    system_report_list_parser = system_report_subcommands.add_parser(
        "list",
        description="Adds reporting code to a given module",
    )
    populate_leaf_parser_with_common_args(system_report_list_parser)
    system_report_list_parser.set_defaults(func=handle_system_report_list)

    system_report_add_parser = system_report_subcommands.add_parser(
        "add",
        description="Adds reporting code to a given module",
    )
    populate_leaf_parser_with_common_args(system_report_add_parser)
    system_report_add_parser.add_argument(
        "-m",
        "--submodule",
        default=None,
        help="Path (relative to Python root) to submodule in which to fire off a system report",
    )
    system_report_add_parser.set_defaults(func=handle_system_report_add)

    system_report_remove_parser = system_report_subcommands.add_parser(
        "remove", description="Removes reporting code from a given module"
    )
    populate_leaf_parser_with_common_args(system_report_remove_parser)
    system_report_remove_parser.add_argument(
        "-m",
        "--submodule",
        default=None,
        help="Path (relative to Python root) to submodule in which to fire off a system report",
    )
    system_report_remove_parser.set_defaults(func=handle_system_report_remove)

    excepthook_parser = subcommands.add_parser(
        "excepthook", description="Manage crash reporting (of all uncaught exceptions)"
    )
    excepthook_parser.set_defaults(func=lambda _: excepthook_parser.print_help())
    excepthook_subcommands = excepthook_parser.add_subparsers()

    (
        handle_excepthook_list,
        handle_excepthook_add,
        handle_excepthook_remove,
    ) = generate_call_handlers(operations.CALL_TYPE_SETUP_EXCEPTHOOK)

    excepthook_list_parser = excepthook_subcommands.add_parser(
        "list",
        description="Adds reporting code to a given module",
    )
    populate_leaf_parser_with_common_args(excepthook_list_parser)
    excepthook_list_parser.set_defaults(func=handle_excepthook_list)

    excepthook_add_parser = excepthook_subcommands.add_parser(
        "add",
        description="Adds crash reporting to a given package",
    )
    populate_leaf_parser_with_common_args(excepthook_add_parser)
    excepthook_add_parser.set_defaults(func=handle_excepthook_add)

    excepthook_remove_parser = excepthook_subcommands.add_parser(
        "remove",
        description="Adds crash reporting to a given package",
    )
    populate_leaf_parser_with_common_args(excepthook_remove_parser)
    excepthook_remove_parser.set_defaults(func=handle_excepthook_remove)

    record_call_parser = subcommands.add_parser(
        "record-call", description="Record every time a function/method is called"
    )
    record_call_parser.set_defaults(func=lambda _: record_call_parser.print_help())
    record_call_subcommands = record_call_parser.add_subparsers()

    (
        handle_record_call_list,
        handle_record_call_candidates,
        handle_record_call_add,
        handle_record_call_remove,
    ) = generate_decorator_handlers(operations.DECORATOR_TYPE_RECORD_CALL)

    record_call_list_parser = record_call_subcommands.add_parser(
        "list",
        description="List all functions/methods which are currently being recorded",
    )
    populate_leaf_parser_with_common_args(record_call_list_parser)
    record_call_list_parser.set_defaults(func=handle_record_call_list)

    record_call_candidates_parser = record_call_subcommands.add_parser(
        "candidates",
        description="List all functions/methods in the given submodule on which we can add the decorator",
    )
    populate_leaf_parser_with_common_args(record_call_candidates_parser)
    record_call_candidates_parser.add_argument(
        "-m",
        "--submodule",
        required=True,
        help="Path (relative to Python root) to submodule in which list candidates",
    )
    record_call_candidates_parser.set_defaults(func=handle_record_call_candidates)

    record_call_add_parser = record_call_subcommands.add_parser(
        "add",
        description="Adds reporting code to a given module",
    )
    populate_leaf_parser_with_common_args(record_call_add_parser)
    record_call_add_parser.add_argument(
        "-m",
        "--submodule",
        required=True,
        help="Path (relative to Python root) to submodule in which list candidates",
    )
    record_call_add_parser.add_argument(
        "lines",
        type=int,
        nargs="+",
        help="Line numbers of function definitions to decorate",
    )
    record_call_add_parser.set_defaults(func=handle_record_call_add)

    record_call_remove_parser = record_call_subcommands.add_parser(
        "remove",
        description="List all functions/methods which are currently being recorded",
    )
    populate_leaf_parser_with_common_args(record_call_remove_parser)
    record_call_remove_parser.add_argument(
        "-m",
        "--submodule",
        required=True,
        help="Path (relative to Python root) to submodule in which list candidates",
    )
    record_call_remove_parser.add_argument(
        "lines",
        type=int,
        nargs="+",
        help="Line numbers of function definitions to decorate",
    )
    record_call_remove_parser.set_defaults(func=handle_record_call_remove)

    record_error_parser = subcommands.add_parser(
        "record-error",
        description="Record function/method's caught and uncaught errors",
    )
    record_error_parser.set_defaults(func=lambda _: record_error_parser.print_help())
    record_error_subcommands = record_error_parser.add_subparsers()

    (
        handle_record_error_list,
        handle_record_error_candidates,
        handle_record_error_add,
        handle_record_error_remove,
    ) = generate_decorator_handlers(operations.DECORATOR_TYPE_RECORD_ERRORS)

    record_error_list_parser = record_error_subcommands.add_parser(
        "list",
        description="List all functions/methods which are currently being recorded",
    )
    populate_leaf_parser_with_common_args(record_error_list_parser)
    record_error_list_parser.set_defaults(func=handle_record_error_list)

    record_error_candidates_parser = record_error_subcommands.add_parser(
        "candidates",
        description="List all functions/methods in the given submodule on which we can add the decorator",
    )
    populate_leaf_parser_with_common_args(record_error_candidates_parser)
    record_error_candidates_parser.add_argument(
        "-m",
        "--submodule",
        required=True,
        help="Path (relative to Python root) to submodule in which list candidates",
    )
    record_error_candidates_parser.set_defaults(func=handle_record_error_candidates)

    record_error_add_parser = record_error_subcommands.add_parser(
        "add",
        description="Adds reporting code to a given module",
    )
    populate_leaf_parser_with_common_args(record_error_add_parser)
    record_error_add_parser.add_argument(
        "-m",
        "--submodule",
        required=True,
        help="Path (relative to Python root) to submodule in which list candidates",
    )
    record_error_add_parser.add_argument(
        "lines",
        type=int,
        nargs="+",
        help="Line numbers of function definitions to decorate",
    )
    record_error_add_parser.set_defaults(func=handle_record_error_add)

    record_error_remove_parser = record_error_subcommands.add_parser(
        "remove",
        description="List all functions/methods which are currently being recorded",
    )
    populate_leaf_parser_with_common_args(record_error_remove_parser)
    record_error_remove_parser.add_argument(
        "-m",
        "--submodule",
        required=True,
        help="Path (relative to Python root) to submodule in which list candidates",
    )
    record_error_remove_parser.add_argument(
        "lines",
        type=int,
        nargs="+",
        help="Line numbers of function definitions to decorate",
    )
    record_error_remove_parser.set_defaults(func=handle_record_error_remove)

    report_all_parser = subcommands.add_parser(
        "report-all", description="Report all that can be reported "
    )
    populate_leaf_parser_with_common_args(report_all_parser)
    report_all_parser.set_defaults(func=handle_report_all)

    return parser


def main() -> None:
    parser = generate_argument_parser()
    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    main()



"""
These are the tools infestor uses to set up a code base for automatic Humbug instrumentation.
"""
from dataclasses import asdict, dataclass
import json
import os
from typing import Any, cast, Dict, List, Optional, Tuple

from atomicwrites import atomic_write

from .version import INFESTOR_VERSION


CONFIG_FILENAME = "humbug.json"


@dataclass
class InfestorConfiguration:
    project_name: str
    relative_imports: bool = False
    reporter_token: Optional[str] = None
    # reporter_filepath should be a path relative to Python root
    reporter_filepath: Optional[str] = None


PROJECT_NAME_KEY = "project_name"
RELATIVE_IMPORTS_KEY = "relative_imports"
REPORTER_TOKEN_KEY = "reporter_token"
REPORTER_FILEPATH_KEY = "reporter_filepath"


class ConfigurationError(Exception):
    """
    Raised if there is an issue with an infestor configuration file.
    """

    pass


def parse_config(
    raw_config: Dict[str, Any]
) -> Tuple[Optional[InfestorConfiguration], List[str], List[str]]:
    """
    Checks if the given configuration is valid. If it is valid, returns (True, []) else returns
    (False, [<warnings>, ...], [<error messages>, ...]).
    """
    warn_messages: List[str] = []
    error_messages: List[str] = []

    infestor_configuration: Optional[InfestorConfiguration] = None

    project_name: Optional[str] = raw_config.get(PROJECT_NAME_KEY)
    if project_name is None:
        error_messages.append("No project name specified")
        project_name = ""

    relative_imports: Optional[bool] = raw_config.get(RELATIVE_IMPORTS_KEY)
    if relative_imports is None:
        error_messages.append(
            "Configuration does not specify whether or not to use relative imports"
        )
        relative_imports = False

    reporter_token: Optional[str] = raw_config.get(REPORTER_TOKEN_KEY)
    if reporter_token is None:
        warn_messages.append(f"No reporter token found")

    reporter_filepath: Optional[str] = raw_config.get(REPORTER_FILEPATH_KEY)
    if reporter_filepath is None:
        warn_messages.append(f"No reporter filepath found")

    if not error_messages:
        infestor_configuration = InfestorConfiguration(
            project_name=project_name,
            relative_imports=relative_imports,
            reporter_token=reporter_token,
            reporter_filepath=reporter_filepath,
        )

    return (infestor_configuration, warn_messages, error_messages)


def load_config(
    config_file: str, print_warnings: bool = False
) -> InfestorConfiguration:
    """
    Loads an infestor configuration from file and validates it.
    """
    try:
        with open(config_file, "r") as ifp:
            raw_config = json.load(ifp)
    except:
        raise ConfigurationError(f"Could not read configuration: {config_file}")

    configuration, warnings, errors = parse_config(raw_config)

    if print_warnings:
        warning_items = "\n".join([f"- {warning}" for warning in warnings])
        if warnings:
            print(
                f"Warnings when loading configuration file ({config_file}):\n{warning_items}"
            )

    if errors:
        error_items = "\n".join([f"- {error}" for error in errors])
        error_message = (
            f"Errors loading configuration file ({config_file}):\n{error_items}"
        )
        raise ConfigurationError(error_message)

    return cast(InfestorConfiguration, configuration)


def save_config(config_file: str, configuration: InfestorConfiguration) -> None:
    result_configuration = asdict(configuration)
    with atomic_write(config_file, overwrite=True) as ofp:
        json.dump(result_configuration, ofp)


def default_config_file(root_directory) -> str:
    config_file = os.path.join(root_directory, CONFIG_FILENAME)
    return config_file


def set_reporter_token(config_file: str, reporter_token: str) -> InfestorConfiguration:
    configuration = load_config(config_file)
    configuration.reporter_token = reporter_token
    save_config(config_file, configuration)
    return configuration


def initialize(
    repository: str,
    project_name: str,
    relative_imports: bool = False,
    reporter_token: Optional[str] = None,
) -> InfestorConfiguration:
    """
    Initialize infestor in a given project.
    """
    config_file = default_config_file(repository)
    configuration = InfestorConfiguration(
        project_name=project_name,
        relative_imports=relative_imports,
        reporter_token=reporter_token,
    )
    save_config(config_file, configuration)
    return configuration


def python_root_relative_to_repository_root(repository: str, python_root: str) -> str:
    absolute_python_root = os.path.abspath(python_root)
    relative_python_root = os.path.relpath(absolute_python_root, repository)
    return relative_python_root



import os
from typing import List

import pygit2


def commit_files(
    repository: str,
    ref: str,
    filepaths: List[str],
    message: str,
    author: str = "Infestor",
    email: str = "infestor@bugout.dev",
) -> str:
    """
    Adds the given files to the repo index and makes a commit.

    Pygit2 commit recipe: https://gist.github.com/lig/dc1ede7e09488a62116fe90aa31617d9
    """
    signature = pygit2.Signature(author, email)
    repo = pygit2.Repository(path=repository)
    for filepath in filepaths:
        repo.index.add(filepath)
    tree = repo.index.write_tree()
    parents = []
    try:
        parent, _ = repo.resolve_refish(refish=repo.head.name)
        parents.append(parent.oid)
    except Exception:
        pass
    commit_oid = repo.create_commit(
        ref,
        signature,
        signature,
        message,
        tree,
        parents,
    )
    return commit_oid



import difflib
import unittest
from typing import Optional

import libcst as cst
import libcst.matchers as m

from . import config
from . import transformers
from .operations import add_reporter
from .testcase import InfestorTestCase


class TestTryExcept(unittest.TestCase):
    def setUp(self) -> None:
        pass

    def test_simple_try_except(self):
        source_code = """
        """


class TestImportReporterTransformer(InfestorTestCase):
    def setUp(self):
        super().setUp()
        add_reporter(self.package_dir)
        self.config = config.load_config(self.config_file)
        self.package_transformer = transformers.ImportReporterTransformer(
            self.package_dir
        )


if __name__ == "__main__":
    unittest.main()



from typing import Optional, List, Tuple, Dict, cast

import libcst.matchers as m
import libcst as cst
import logging
from . import models
from . import manager
from . import transformers


class ReporterNotImported(Exception):
    pass


class ReporterFileVisitor(cst.CSTVisitor):
    METADATA_DEPENDENCIES = (cst.metadata.PositionProvider,)

    def __init__(self):
        self.HumbugConsentImportedAs: str = ""
        self.HumbugConsentImportedAt: int = -1
        self.HumbugReporterImportedAs: str = ""
        self.HumbugReporterImportedAt: int = -1
        self.HumbugConsentInstantiatedAt: int = -1
        self.HumbugConsentInstantiatedAs: str = ""
        self.HumbugReporterInstantiatedAs: str = ""
        self.HumbugReporterInstantiatedAt: int = -1
        self.HumbugReporterConsentArgument: str = ""
        self.HumbugReporterTokenArgument: str = ""

    @staticmethod
    def syntax_tree(reporter_filepath: str) -> cst.MetadataWrapper:
        with open(reporter_filepath, "r") as ifp:
            reporter_file_source = ifp.read()
        reporter_syntax_tree = cst.metadata.MetadataWrapper(
            cst.parse_module(reporter_file_source)
        )
        return reporter_syntax_tree

    def visit_ImportFrom(self, node: cst.ImportFrom) -> Optional[bool]:
        position = self.get_metadata(cst.metadata.PositionProvider, node)  # type: ignore
        if (
            isinstance(node.module, cst.Attribute)
            and isinstance(node.module.value, cst.Name)
            and node.module.value.value == "humbug"
        ):
            if node.module.attr.value == "consent" and not isinstance(
                node.names, cst.ImportStar
            ):
                for name in node.names:
                    if name.name.value == "HumbugConsent":
                        self.HumbugConsentImportedAs = "HumbugConsent"

                        if name.asname is not None and isinstance(
                            name.asname, cst.Name
                        ):
                            self.HumbugConsentImportedAs = name.asname.value

                        self.HumbugConsentImportedAt = position.start.line
            elif node.module.attr.value == "report" and not isinstance(
                node.names, cst.ImportStar
            ):
                for name in node.names:
                    if name.name.value == "HumbugReporter":
                        self.HumbugReporterImportedAs = "HumbugReporter"

                        if name.asname is not None and isinstance(
                            name.asname, cst.Name
                        ):
                            self.HumbugReporterImportedAs = name.asname.value

                        self.HumbugReporterImportedAt = position.start.line

        return False

    def visit_Assign(self, node: cst.Assign) -> Optional[bool]:
        if (
            len(node.targets) == 1
            and isinstance(node.value, cst.Call)
            and isinstance(node.value.func, cst.Name)
            and isinstance(node.targets[0].target, cst.Name)
        ):
            if node.value.func.value == self.HumbugConsentImportedAs:
                position = self.get_metadata(cst.metadata.PositionProvider, node)  # type: ignore
                self.HumbugConsentInstantiatedAt = position.start.line
                self.HumbugConsentInstantiatedAs = node.targets[0].target.value
                return False
            elif node.value.func.value == self.HumbugReporterImportedAs:
                position = self.get_metadata(cst.metadata.PositionProvider, node)  # type: ignore
                self.HumbugReporterInstantiatedAt = position.start.line
                self.HumbugReporterInstantiatedAs = node.targets[0].target.value
        return True

    def visit_Call(self, node: cst.Call) -> Optional[bool]:
        if (
            isinstance(node.func, cst.Name)
            and node.func.value == self.HumbugReporterImportedAs
        ):
            for arg in node.args:
                if (
                    arg.keyword is not None
                    and arg.keyword.value == "consent"
                    and isinstance(arg.value, cst.Name)
                ):
                    self.HumbugReporterConsentArgument = arg.value.value
                elif (
                    arg.keyword is not None
                    and arg.keyword.value == "bugout_token"
                    and isinstance(arg.value, cst.SimpleString)
                ):
                    self.HumbugReporterTokenArgument = arg.value.value
        return False


class DecoratorCandidatesVisitor(cst.CSTVisitor):
    METADATA_DEPENDENCIES = (cst.metadata.PositionProvider,)

    def __init__(self, reporter_imported_as, decorator_type):
        self.reporter_imported_as = reporter_imported_as
        self.decorator_type = decorator_type
        self.scope_stack: List[str] = []
        self.decorator_candidates: List[models.ReporterDecoratorCandidate] = []

    def visit_FunctionDef(self, node: cst.FunctionDef) -> Optional[bool]:
        self.scope_stack.append(node.name.value)

        for decorator in node.decorators:
            if transformers.matches_with_reporter_decorator(
                decorator, self.reporter_imported_as, self.decorator_type
            ):
                return True

        position = self.get_metadata(cst.metadata.PositionProvider, node)
        self.decorator_candidates.append(
            models.ReporterDecoratorCandidate(
                scope_stack=".".join(self.scope_stack), lineno=position.start.line
            )
        )
        return True

    def leave_FunctionDef(self, original_node: cst.FunctionDef) -> None:
        self.scope_stack.pop()

    def visit_ClassDef(self, node: cst.ClassDef) -> Optional[bool]:
        self.scope_stack.append(node.name.value)
        return True

    def leave_ClassDef(self, original_node: cst.ClassDef) -> None:
        self.scope_stack.pop()


class PackageFileVisitor(cst.CSTVisitor):
    METADATA_DEPENDENCIES = (cst.metadata.PositionProvider,)
    last_import_lineno = 0

    def __init__(self, reporter_module_path: str, relative_imports: bool):
        self.ReporterImportedAs: str = ""
        self.ReporterImportedAt: int = -1
        self.ReporterCorrectlyImported: bool = False

        self.relative_imports = relative_imports
        self.reporter_module_path = reporter_module_path
        self.scope_stack: List[str] = []

        self.calls: Dict[str, List[models.ReporterCall]] = {}
        self.decorators: Dict[str, List[models.ReporterDecorator]] = {}

    # TODO(yhtiyar) also add checking with 'import'
    def matches_with_package_import(self, node: cst.ImportFrom):
        return m.matches(
            node,
            m.ImportFrom(
                module=m.Attribute(
                    value=m.Name(
                        # TODO: Refactor this
                        value=self.reporter_module_path.rsplit(".", 1)[
                            0
                        ]  # checking for reporter module path basename
                    ),
                    attr=m.Name(value="report"),
                ),
            ),
        )

    def matches_reporter_call(self, node: cst.Call):
        return m.matches(
            node,
            m.Call(
                func=m.Attribute(
                    value=m.Name(value=self.ReporterImportedAs),
                ),
            ),
        )

    def matches_with_reporter_decorator(self, node: cst.Decorator):
        return m.matches(
            node,
            m.Decorator(
                decorator=m.Attribute(
                    value=m.Name(value=self.ReporterImportedAs),
                )
            ),
        )

    def visit_FunctionDef(self, node: cst.FunctionDef):
        self.scope_stack.append(node.name.value)
        for decorator in node.decorators:
            if self.matches_with_reporter_decorator(decorator):
                position = self.get_metadata(cst.metadata.PositionProvider, node)
                decorator_attribute = cast(cst.Attribute, decorator.decorator)
                decorator_model = models.ReporterDecorator(
                    decorator_type=decorator_attribute.attr.value,
                    scope_stack=".".join(self.scope_stack),
                    lineno=position.start.line,
                )
                self.decorators.setdefault(decorator_model.decorator_type, []).append(
                    decorator_model
                )
        return True

    def leave_FunctionDef(self, original_node: cst.FunctionDef) -> None:
        self.scope_stack.pop()

    def visit_ClassDef(self, node: cst.ClassDef):
        self.scope_stack.append(node.name.value)
        return True

    def leave_ClassDef(self, original_node: cst.ClassDef) -> None:
        self.scope_stack.pop()

    def visit_Import(self, node: cst.Import) -> Optional[bool]:
        if self.scope_stack:
            return False
        position = self.get_metadata(cst.metadata.PositionProvider, node)
        self.last_import_lineno = position.end.line
        return False

    def check_alias_for_reporter(self, import_aliases, position):
        for alias in import_aliases:
            name = alias.name.value
            if name == "reporter":
                asname = alias.asname
                if asname:
                    self.ReporterImportedAs = asname.name.value
                else:
                    self.ReporterImportedAs = name
                self.ReporterImportedAt = position.start.line
                self.ReporterCorrectlyImported = (
                    position.start.line == self.last_import_lineno + 1
                )

    def visit_ImportFrom(self, node: cst.ImportFrom) -> Optional[bool]:
        if self.scope_stack:
            return False
        position = self.get_metadata(cst.metadata.PositionProvider, node)
        if self.relative_imports:
            expected_level = 0
            for character in self.reporter_module_path:
                if character == ".":
                    expected_level += 1
                else:
                    break
            node_import_level = 0
            for dot in node.relative:
                if isinstance(dot, cst.Dot):
                    node_import_level += 1
                else:
                    break
            if (
                node_import_level == expected_level
                and isinstance(node.module, cst.Name)
                and node.module.value == self.reporter_module_path[expected_level:]
            ):
                import_aliases = node.names
                self.check_alias_for_reporter(import_aliases, position)

        elif self.matches_with_package_import(node):
            import_aliases = node.names
            self.check_alias_for_reporter(import_aliases, position)

        self.last_import_lineno = position.end.line
        return False

    def visit_Call(self, node: cst.Call) -> Optional[bool]:
        if self.ReporterImportedAt == -1:
            return False
        if self.matches_reporter_call(node):
            position = self.get_metadata(cst.metadata.PositionProvider, node)
            func_attr = cast(cst.Attribute, node.func)
            call_model = models.ReporterCall(
                call_type=func_attr.attr.value,
                lineno=position.start.line,
                scope_stack=".".join(self.scope_stack),
            )
            self.calls.setdefault(call_model.call_type, []).append(call_model)
        return False



from typing import Optional, List, Union, cast

import libcst as cst
import libcst.matchers as m


def matches_import(node: cst.CSTNode) -> bool:
    return m.matches(
        node,
        m.SimpleStatementLine(
            body=m.MatchIfTrue(
                lambda body: all(m.matches(el, m.Import | m.ImportFrom) for el in body)
            )
        ),
    )


class ImportReporterTransformer(cst.CSTTransformer):
    """
    Imports reporter from reporter_module_path path after last naked import
    """

    def __init__(self, reporter_module_path):
        self.reporter_import_code = f"from {reporter_module_path} import reporter"
        self.last_import: Optional[cst.CSTNode] = None

    def visit_Module(self, node: cst.Module) -> Optional[bool]:
        for statement in node.body:
            if matches_import(statement):
                self.last_import = statement
        return False

    def leave_Module(
        self, original_node: cst.Module, updated_node: cst.Module
    ) -> cst.Module:
        new_body: List[cst.CSTNode] = []

        parsed_reporter_import = cst.parse_statement(
            self.reporter_import_code, original_node.config_for_parsing
        )

        if self.last_import is None:
            new_body.append(parsed_reporter_import)

        for el in original_node.body:
            new_body.append(el)
            if self.last_import and el == self.last_import:
                new_body.append(parsed_reporter_import)

        return updated_node.with_changes(body=new_body)


class ReporterCallsAdderTransformer(cst.CSTTransformer):
    def __init__(self, reporter_imported_as: str, call_type: str):
        # self.call_to_add = cst.parse_statement(f"{reporter_imported_as}.{call_type}()")
        self.call_to_add = cst.SimpleStatementLine(
            body=[
                cst.Expr(
                    value=cst.Call(
                        func=cst.Attribute(
                            value=cst.Name(value=reporter_imported_as),
                            attr=cst.Name(value=call_type),
                        )
                    )
                )
            ]
        )
        self.last_import: Optional[cst.CSTNode] = None

    def visit_Module(self, node: cst.Module) -> Optional[bool]:
        for statement in node.body:
            if matches_import(statement):
                self.last_import = statement
        return False

    def leave_Module(
        self, original_node: cst.Module, updated_node: cst.Module
    ) -> cst.Module:
        new_body = []

        if self.last_import is None:
            raise Exception("No import found when tried to add call")

        for el in original_node.body:
            new_body.append(el)
            if el == self.last_import:
                new_body.append(self.call_to_add)

        return updated_node.with_changes(body=new_body)


class ReporterCallsRemoverTransformer(cst.CSTTransformer):
    def matches_reporter_call(self, node: cst.CSTNode):
        return m.matches(
            node,
            m.SimpleStatementLine(
                body=[
                    m.Expr(
                        value=m.Call(
                            func=m.Attribute(
                                value=m.Name(value=self.reporter_imported_as),
                                attr=m.Name(value=self.call_type),
                            ),
                        ),
                    )
                ]
            ),
        )

    def __init__(self, reporter_imported_as: str, call_type: str):

        self.reporter_imported_as = reporter_imported_as
        self.call_type = call_type

    def leave_Module(
        self, original_node: cst.Module, updated_node: cst.Module
    ) -> cst.Module:
        new_body = []

        for el in original_node.body:
            if not self.matches_reporter_call(el):
                new_body.append(el)

        return updated_node.with_changes(body=new_body)


ERROR_REPORT_CALL = "error_report"


class TryExceptAdderTransformer(cst.CSTTransformer):
    METADATA_DEPENDENCIES = (cst.metadata.PositionProvider,)

    def __init__(self, reported_imported_as: str, linenos: List[int]):
        self.reporter_imported_as = reported_imported_as
        self.linenos = linenos
        self.func_scope: List[int] = []

    def has_except_asname(self, node: cst.ExceptHandler):
        return m.matches(node, m.ExceptHandler(name=m.AsName(name=m.Name())))

    def matches_error_report_call(self, node: cst.CSTNode, except_as_name):
        return m.matches(
            node,
            m.Call(
                func=m.Attribute(
                    value=m.Name(value=self.reporter_imported_as),
                    attr=m.Name(value=ERROR_REPORT_CALL),
                ),
                args=[m.Arg(value=m.Name(value=except_as_name))],
            ),
        )

    def matches_error_report_statement(
        self, node: cst.SimpleStatementLine, except_as_name
    ):
        return m.matches(
            node,
            m.SimpleStatementLine(
                body=[
                    m.Expr(
                        value=m.MatchIfTrue(
                            lambda value: self.matches_error_report_call(
                                value, except_as_name
                            )
                        )
                    )
                ]
            ),
        )

    def visit_FunctionDef(self, node: cst.FunctionDef) -> Optional[bool]:
        position = self.get_metadata(cst.metadata.PositionProvider, node)
        self.func_scope.append(position.start.line)
        return True

    def leave_FunctionDef(
        self, original_node: cst.FunctionDef, updated_node: cst.FunctionDef
    ) -> Union[
        cst.BaseStatement, cst.FlattenSentinel[cst.BaseStatement], cst.RemovalSentinel
    ]:
        self.func_scope.pop()
        return updated_node

    def leave_ExceptHandler(
        self, node: cst.ExceptHandler, updated_node: cst.ExceptHandler
    ) -> Union[
        cst.ExceptHandler, cst.FlattenSentinel[cst.ExceptHandler], cst.RemovalSentinel
    ]:
        if not self.func_scope:
            return updated_node
        if self.func_scope[-1] not in self.linenos:
            return updated_node

        asname = "e"
        new_name = node.name
        except_type = node.type

        if except_type is None:
            except_type = cst.Name(value="BaseException")

        if self.has_except_asname(node):
            # Casting to types, since getting errors in mypy
            _asname = cast(cst.AsName, node.name)
            name = cast(cst.Name, _asname.name)
            asname = name.value
        else:
            new_name = cst.AsName(name=cst.Name(value=asname))

        new_inner_body = []
        has_called_error_report = False
        for (
            el
        ) in (
            updated_node.body.body
        ):  # Using updated node, since child od node is updated
            new_inner_body.append(el)
            if isinstance(
                el, cst.SimpleStatementLine
            ) and self.matches_error_report_statement(el, asname):
                has_called_error_report = True

        if not has_called_error_report:
            new_inner_body.insert(
                0,
                cst.parse_statement(
                    f"{self.reporter_imported_as}.{ERROR_REPORT_CALL}({asname})"
                ),
            )
        new_body = updated_node.body.with_changes(body=new_inner_body)

        return updated_node.with_changes(
            name=new_name,
            type=except_type,
            body=new_body,
            whitespace_after_except=cst.SimpleWhitespace(
                value=" ",
            ),
        )


class TryExceptRemoverTransformer(cst.CSTTransformer):
    METADATA_DEPENDENCIES = (cst.metadata.PositionProvider,)

    def __init__(self, reported_imported_as: str, linenos: List[int]):
        self.reporter_imported_as = reported_imported_as
        self.linenos = linenos
        self.func_scope: List[int] = []

    def has_except_asname(self, node: cst.ExceptHandler):
        return m.matches(node, m.ExceptHandler(name=m.AsName(name=m.Name())))

    def matches_error_report_call(self, node: cst.CSTNode, except_as_name):
        return m.matches(
            node,
            m.Call(
                func=m.Attribute(
                    value=m.Name(value=self.reporter_imported_as),
                    attr=m.Name(value=ERROR_REPORT_CALL),
                ),
                args=[m.Arg(value=m.Name(value=except_as_name))],
            ),
        )

    def matches_error_report_statement(
        self, node: cst.SimpleStatementLine, except_as_name
    ):
        return m.matches(
            node,
            m.SimpleStatementLine(
                body=[
                    m.Expr(
                        value=m.MatchIfTrue(
                            lambda value: self.matches_error_report_call(
                                value, except_as_name
                            )
                        )
                    )
                ]
            ),
        )

    def visit_FunctionDef(self, node: cst.FunctionDef) -> Optional[bool]:
        position = self.get_metadata(cst.metadata.PositionProvider, node)
        self.func_scope.append(position.start.line)
        return True

    def leave_FunctionDef(
        self, original_node: cst.FunctionDef, updated_node: cst.FunctionDef
    ) -> Union[
        cst.BaseStatement, cst.FlattenSentinel[cst.BaseStatement], cst.RemovalSentinel
    ]:
        self.func_scope.pop()
        return updated_node

    def leave_ExceptHandler(
        self, node: cst.ExceptHandler, updated_node: cst.ExceptHandler
    ) -> Union[
        cst.ExceptHandler, cst.FlattenSentinel[cst.ExceptHandler], cst.RemovalSentinel
    ]:
        if not self.func_scope:
            return updated_node
        if self.func_scope[-1] not in self.linenos:
            return updated_node

        asname = "e"
        new_name = node.name
        except_type = node.type

        if except_type is None:
            return updated_node

        if self.has_except_asname(node):
            # Casting to types, since getting errors in mypy
            _asname = cast(cst.AsName, node.name)
            name = cast(cst.Name, _asname.name)
            asname = name.value
        else:
            return updated_node

        new_inner_body = []

        for (
            el
        ) in (
            updated_node.body.body
        ):  # Using updated node, since child od node is updated

            if not (
                isinstance(el, cst.SimpleStatementLine)
                and self.matches_error_report_statement(el, asname)
            ):
                new_inner_body.append(el)

        new_body = updated_node.body.with_changes(body=new_inner_body)

        return updated_node.with_changes(
            body=new_body,
        )


def matches_with_reporter_decorator(
    node: cst.Decorator, reporter_imported_as, decorator_type
):
    return m.matches(
        node,
        m.Decorator(
            decorator=m.Attribute(
                value=m.Name(value=reporter_imported_as),
                attr=m.Name(value=decorator_type),
            )
        ),
    )


class DecoratorsAdderTransformer(cst.CSTTransformer):
    METADATA_DEPENDENCIES = (cst.metadata.PositionProvider,)

    def __init__(self, reporter_imported_as, decorator_type, lines_to_add: List[int]):
        self.reporter_imported_as = reporter_imported_as
        self.lines_to_add = lines_to_add
        self.decorator_type = decorator_type
        self.decorator_to_add = cst.Decorator(
            decorator=cst.Attribute(
                value=cst.Name(
                    value=reporter_imported_as,
                ),
                attr=cst.Name(
                    value=decorator_type,
                ),
            )
        )

    def leave_FunctionDef(self, original_node, updated_node):
        position = self.get_metadata(cst.metadata.PositionProvider, original_node)
        if position.start.line not in self.lines_to_add:
            return updated_node

        decorators = [self.decorator_to_add]
        for decorator in updated_node.decorators:
            decorators.append(decorator)
            if matches_with_reporter_decorator(
                decorator, self.reporter_imported_as, self.decorator_type
            ):
                return updated_node

        return updated_node.with_changes(decorators=decorators)


class DecoratorsRemoverTransformer(cst.CSTTransformer):
    METADATA_DEPENDENCIES = (cst.metadata.PositionProvider,)

    def __init__(
        self, reporter_imported_as, decorator_type, lines_to_remove: List[int]
    ):
        self.reporter_imported_as = reporter_imported_as
        self.decorator_type = decorator_type
        self.lines_to_remove = lines_to_remove

    def leave_FunctionDef(self, original_node, updated_node):
        position = self.get_metadata(cst.metadata.PositionProvider, original_node)

        if position.start.line not in self.lines_to_remove:
            return updated_node

        decorators = []
        for decorator in updated_node.decorators:
            if not matches_with_reporter_decorator(
                decorator, self.reporter_imported_as, self.decorator_type
            ):
                decorators.append(decorator)

        return updated_node.with_changes(decorators=decorators)



from dataclasses import asdict
import json
import os
import shutil
import tempfile
import unittest

from . import config


class TestParseConfig(unittest.TestCase):
    maxDiff = None

    # TODO(zomglings): Test reporter_filepath invalidities.

    def test_empty_config(self):
        raw_config = {}
        configuration, warnings, errors = config.parse_config(raw_config)
        self.assertEqual(len(warnings), 2)
        self.assertEqual(len(errors), 2)
        self.assertIsNone(configuration)

    def test_config_with_no_project_name(self):
        raw_config = {
            config.RELATIVE_IMPORTS_KEY: False,
            config.REPORTER_TOKEN_KEY: "335da960-2dc6-48b3-97a5-c23ac1495e7d",
            config.REPORTER_FILEPATH_KEY: "report.py",
        }
        configuration, warnings, errors = config.parse_config(raw_config)
        self.assertEqual(len(warnings), 0)
        self.assertEqual(len(errors), 1)
        self.assertIsNone(configuration)

    def test_config_with_no_relative_imports(self):
        raw_config = {
            config.PROJECT_NAME_KEY: "rofl",
            config.REPORTER_TOKEN_KEY: "335da960-2dc6-48b3-97a5-c23ac1495e7d",
            config.REPORTER_FILEPATH_KEY: "report.py",
        }
        configuration, warnings, errors = config.parse_config(raw_config)
        self.assertEqual(len(warnings), 0)
        self.assertEqual(len(errors), 1)
        self.assertIsNone(configuration)

    def test_valid_config(self):
        infestor_configuration = config.InfestorConfiguration(
            project_name="lol",
            relative_imports=False,
            reporter_token="335da960-2dc6-48b3-97a5-c23ac1495e7d",
            reporter_filepath="report.py",
        )
        raw_config = {
            config.PROJECT_NAME_KEY: infestor_configuration.project_name,
            config.RELATIVE_IMPORTS_KEY: infestor_configuration.relative_imports,
            config.REPORTER_TOKEN_KEY: infestor_configuration.reporter_token,
            config.REPORTER_FILEPATH_KEY: infestor_configuration.reporter_filepath,
        }
        configuration, warnings, errors = config.parse_config(raw_config)
        self.assertEqual(len(warnings), 0)
        self.assertEqual(len(errors), 0)
        self.assertEqual(configuration, infestor_configuration)


class TestInit(unittest.TestCase):
    def setUp(self):
        self.repository = tempfile.mkdtemp()
        self.project_name = "my-awesome-python-project"

    def tearDown(self):
        shutil.rmtree(self.repository)

    def test_initialize_infestor_once(self):
        configuration = config.initialize(self.repository, self.project_name)
        config_file = config.default_config_file(self.repository)
        self.assertTrue(os.path.isfile(config_file))
        with open(config_file, "r") as ifp:
            configuration_json = json.load(ifp)
        self.assertDictEqual(configuration_json, asdict(configuration))

    def test_initialize_infestor_twice(self):
        initial_configuration = config.initialize(
            self.repository, "lol", relative_imports=False
        )
        final_configuration = config.initialize(
            self.repository, "lol", relative_imports=True
        )

        config_file = config.default_config_file(self.repository)
        self.assertTrue(os.path.isfile(config_file))
        with open(config_file, "r") as ifp:
            configuration_json = json.load(ifp)

        self.assertDictEqual(configuration_json, asdict(final_configuration))


if __name__ == "__main__":
    unittest.main()
